print("import: 'nuitka'")
import nuitka

print("import: 'nuitka.build'")
import nuitka.build

print("import: 'nuitka.containers'")
import nuitka.containers

print("import: 'nuitka.finalizations'")
import nuitka.finalizations

print("import: 'nuitka.freezer'")
import nuitka.freezer

print("import: 'nuitka.importing'")
import nuitka.importing

print("import: 'nuitka.nodes'")
import nuitka.nodes

print("import: 'nuitka.optimizations'")
import nuitka.optimizations

print("import: 'nuitka.plugins'")
import nuitka.plugins

print("import: 'nuitka.specs'")
import nuitka.specs

print("import: 'nuitka.tools'")
import nuitka.tools

print("import: 'nuitka.tree'")
import nuitka.tree

print("import: 'nuitka.utils'")
import nuitka.utils

